@@ -5,6 +5,7 @@
 #include "xs_io.h"
 #include "xs_json.h"
 #include "xs_time.h"
+#include "xs_openssl.h"
 
 #include "snac.h"
 
@@ -30,7 +31,8 @@ int usage(void)
     printf("unfollow {basedir} {uid} {actor}    Unfollows an actor\n");
     printf("request {basedir} {uid} {url}       Requests an object\n");
     printf("actor {basedir} [{uid}] {url}       Requests an actor\n");
-    printf("note {basedir} {uid} {'text'}       Sends a note to followers\n");
+    printf("note {basedir} {uid} {text} [files...] Sends a note with optional attachments\n");
+    printf("announce {basedir} {uid} {url}      Announces (boosts) a post\n");
     printf("resetpwd {basedir} {uid}            Resets the password of a user\n");
     printf("ping {basedir} {uid} {actor}        Pings an actor\n");
     printf("webfinger_s {basedir} {uid} {actor} Queries about an actor (@user@host or actor url)\n");
@@ -40,8 +42,7 @@ int usage(void)
     printf("unblock {basedir} {instance_url}    Unblocks a full instance\n");
     printf("limit {basedir} {uid} {actor}       Limits an actor (drops their announces)\n");
     printf("unlimit {basedir} {uid} {actor}     Unlimits an actor\n");
-
-/*    printf("question {basedir} {uid} 'opts'  Generates a poll (;-separated opts)\n");*/
+    printf("verify_links {basedir} {uid}        Verifies a user's links (in the metadata)\n");
 
     return 1;
 }
@@ -58,16 +59,6 @@ char *get_argv(int *argi, int argc, char *argv[])
 
 #define GET_ARGV() get_argv(&argi, argc, argv)
 
-#include "xs_html.h"
-
-xs_html *html_note(snac *user, char *summary,
-                   char *div_id, char *form_id,
-                   char *ta_plh, char *ta_content,
-                   char *edit_id, char *actor_id,
-                   xs_val *cw_yn, char *cw_text,
-                   xs_val *mnt_only, char *redir,
-                   char *in_reply_to, int poll);
-
 int main(int argc, char *argv[])
 {
     char *cmd;
@@ -107,7 +98,7 @@ int main(int argc, char *argv[])
     if (strcmp(cmd, "markdown") == 0) { /** **/
         /* undocumented, for testing only */
         xs *c = xs_readall(stdin);
-        xs *fc = not_really_markdown(c, NULL);
+        xs *fc = not_really_markdown(c, NULL, NULL);
 
         printf("<html>\n%s\n</html>\n", fc);
         return 0;
@@ -247,6 +238,11 @@ int main(int argc, char *argv[])
         return 0;
     }
 
+    if (strcmp(cmd, "verify_links") == 0) { /** **/
+        verify_links(&snac);
+        return 0;
+    }
+
     if (strcmp(cmd, "timeline") == 0) { /** **/
 #if 0
         xs *list = local_list(&snac, XS_ALL);
@@ -366,6 +362,14 @@ int main(int argc, char *argv[])
     if (strcmp(cmd, "ping") == 0) { /** **/
         xs *actor_o = NULL;
 
+        if (!xs_startswith(url, "https:/")) {
+            /* try to resolve via webfinger */
+            if (!valid_status(webfinger_request(url, &url, NULL))) {
+                srv_log(xs_fmt("cannot resolve %s via webfinger", url));
+                return 1;
+            }
+        }
+
         if (valid_status(actor_request(&snac, url, &actor_o))) {
             xs *msg = msg_ping(&snac, url);
 
@@ -374,6 +378,8 @@ int main(int argc, char *argv[])
             if (dbglevel) {
                 xs_json_dump(msg, 4, stdout);
             }
+
+            srv_log(xs_fmt("Ping sent to %s -- see log for Pong reply", url));
         }
         else {
             srv_log(xs_fmt("Error getting actor %s", url));
@@ -456,7 +462,39 @@ int main(int argc, char *argv[])
         xs *content = NULL;
         xs *msg = NULL;
         xs *c_msg = NULL;
-        char *in_reply_to = GET_ARGV();
+        xs *attl = xs_list_new();
+        char *fn = NULL;
+
+        /* iterate possible attachments */
+        while ((fn = GET_ARGV())) {
+            FILE *f;
+
+            if ((f = fopen(fn, "rb")) != NULL) {
+                /* get the file size and content */
+                fseek(f, 0, SEEK_END);
+                int sz = ftell(f);
+                fseek(f, 0, SEEK_SET);
+                xs *atc = xs_readall(f);
+                fclose(f);
+
+                char *ext = strrchr(fn, '.');
+                xs *hash  = xs_md5_hex(fn, strlen(fn));
+                xs *id    = xs_fmt("%s%s", hash, ext);
+                xs *url   = xs_fmt("%s/s/%s", snac.actor, id);
+
+                /* store */
+                static_put(&snac, id, atc, sz);
+
+                xs *l = xs_list_new();
+
+                l = xs_list_append(l, url);
+                l = xs_list_append(l, ""); /* alt text */
+
+                attl = xs_list_append(attl, l);
+            }
+            else
+                fprintf(stderr, "Error opening '%s' as attachment\n", fn);
+        }
 
         if (strcmp(url, "-e") == 0) {
             /* get the content from an editor */
@@ -484,7 +522,7 @@ int main(int argc, char *argv[])
         else
             content = xs_dup(url);
 
-        msg = msg_note(&snac, content, NULL, in_reply_to, NULL, 0);
+        msg = msg_note(&snac, content, NULL, NULL, attl, 0);
 
         c_msg = msg_create(&snac, msg);
 
